# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Zeitschrift Media Perspektiven, refs #11424, #11499.

Config
------

[zmp]

input = /path/to/excel.file.xls
"""

import luigi
from gluish.utils import shellout

from siskin.task import DefaultTask


class ZMPTask(DefaultTask):
    """ Base task for ZMP. """
    TAG = "99"


class ZMPSpreadsheet(ZMPTask, luigi.ExternalTask):
    """ It starts with Excel. """

    def output(self):
        return luigi.LocalTarget(path=self.config.get('zmp', 'input'))


class ZMPMARC(ZMPTask):
    """ Convert Excel to BinaryMarc """

    def requires(self):
        return ZMPSpreadsheet()

    def run(self):
        output = shellout("""python {script} {input} {output}""",
                          script=self.assets("99/99_marcbinary.py"),
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincmarc.mrc'))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301
# Copyright 2018 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
OADOI dataset.


Config
------

[oadoi]

dump = /path/to/full_dois_2018-03-29T113154.jsonl

----

Priem, Jason and Piwowar, Heather. (2018) The Unpaywall Dataset. DOI:
https://doi.org/10.6084/m9.figshare.6020078

> File description
> name: full_dois_2018-03-29T113154.jsonl
> description: This file has one row per DOI, with the best OA location for every DOI, as well as some metadata for each DOI.
> size: 16GB (83G unzipped)
> lines: 95,842,233 lines 
> lines with at least one Open Access URL: ~16 million

Sample row:

    {
        "doi": "10.4135/9781412950534.n2123",
        "year": null,
        "genre": "reference-entry",
        "is_oa": false,
        "title": "Sport Skill Training",
        "doi_url": "https://doi.org/10.4135/9781412950534.n2123",
        "updated": "2018-03-27T15:17:42.869680",
        "publisher": "SAGE Publications, Inc.",
        "z_authors": [
            {
                "given": "David",
                "family": "Reitman"
            },
            {
                "given": "Stephen D. A.",
                "family": "Hupp"
            },
            {
                "given": "Patrick M.",
                "family": "O'Callaghan"
            }
        ],
        "journal_name": "Encyclopedia of Behavior Modification and Cognitive ...",
        "oa_locations": [],
        "data_standard": 1,
        "journal_is_oa": false,
        "journal_issns": null,
        "published_date": null,
        "best_oa_location": null,
        "journal_is_in_doaj": false,
        "x_reported_noncompliant_copies": []
    }

"""

import luigi
from luigi.format import Gzip
from gluish.utils import shellout
from siskin.task import DefaultTask


class OADOITask(DefaultTask):
    """
    Base task for oadoi.
    """
    TAG = 'oadoi'


class OADOIDump(OADOITask, luigi.ExternalTask):
    """
    Provided data dump. See also release notes at:

    * https://docs.google.com/document/d/1Whfe26oyjTedeW1GGWkq3NADgDbL2R2eXqrJCWS8vcc/edit#
    """

    def output(self):
        return luigi.LocalTarget(path=self.config.get("oadoi", "dump"), format=Gzip)


class OADOIList(OADOIDump):
    """
    Provide a simple CSV (doi, is_oa).
    """

    def requires(self):
        return OADOIDump()

    def run(self):
        """
        XXX: parse error: Invalid numeric literal at line 12536737, column 0
        """
        output = shellout("unpigz -c {input} | jq -rc '[.doi, .is_oa] | @csv' | pigz -c > {output}",
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="csv.gz"), format=Gzip)



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
arxiv.org

http://arxiv.org/help/oa/index

----

Config:

[core]

metha-dir = /path/to/dir

----

From: https://groups.google.com/forum/#!topic/arxiv-api/aOacIt6KD2E

> the problem is that the API is not the appropriate vehicle to obtain a full
copy of arXiv metadata. It is intended for interactive queries and search
overlays and should not be used for full replication. The size of result sets
is capped at a large but finite number. You should use OAI-PMH instead for your
purposes.

"""

import datetime

import luigi
from gluish.common import Executable
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.task import DefaultTask


class ArxivTask(DefaultTask):
    """
    Base task.
    """
    TAG = '121'

    def closest(self):
        return weekly(date=self.date)


class ArxivCombine(ArxivTask):
    """
    Single file dump.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    url = luigi.Parameter(default="http://export.arxiv.org/oai2", significant=False)
    prefix = luigi.Parameter(default="oai_dc", significant=False)

    def requires(self):
        return Executable(name='metha-sync', message='https://github.com/miku/metha')

    def run(self):
        shellout("METHA_DIR={dir} metha-sync -format {prefix} {url}",
                 prefix=self.prefix, url=self.url, dir=self.config.get('core', 'metha-dir'))
        output = shellout("METHA_DIR={dir} metha-cat -root Records -format {prefix} {url} | pigz -c > {output}",
                          prefix=self.prefix, url=self.url, dir=self.config.get('core', 'metha-dir'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml.gz"))


class ArxivIntermediateSchema(ArxivTask):
    """
    Better, still experimental.

    Convert to intermediate schema via metafacture. Custom morphs and flux are
    kept in assets/arxiv. Maps are kept in assets/maps.

    Note: There are early arxiv XML records, that do not get an ID in the transformation, e.g.

    {
      "finc.mega_collection": "Arxiv",
      "finc.source_id": "arxiv",
      "rft.atitle": "A determinant of Stirling cycle numbers counts unlabeled acyclic   single-source automata",
      "abstract": "Comment: 11 pages",
      "rft.date": "2007-03-30",
      "x.date": "2007-03-30T00:00:00Z",
      "authors": [
        {
          "rft.au": "Callan, David"
        }
      ],
      "url": [
        "http://arxiv.org/abs/0704.0004"
      ],
      "finc.format": "text",
      "x.subjects": [
        "Mathematics - Combinatorics",
        "05A15"
      ]
    }

    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return ArxivCombine(date=self.date)

    def run(self):
        """
        With 3.1.1 -Xmx512m seemed to be enough. Also: java -Xmx1G -XX:+PrintFlagsFinal -Xmx2G 2>/dev/null | grep MaxHeapSize.
        """
        mapdir = 'file:///%s' % self.assets("maps/")
        output = shellout("""FLUX_JAVA_OPTIONS="-Xmx2048m" flux.sh {flux} in={input} MAP_DIR={mapdir} | pigz -c > {output}""",
                          flux=self.assets("arxiv/arxiv.flux"), mapdir=mapdir, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class ArxivExport(ArxivTask):
    """
    Export to various formats
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3', description='export format')

    def requires(self):
        return ArxivIntermediateSchema(date=self.date)

    def run(self):
        output = shellout("span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}",
                          format=self.format, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2018 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""

Cambridge University Press Journals

refs #10167.
----

Config:

[cambridge]

XXX

"""

from siskin.task import DefaultTask


class CambridgeTask(DefaultTask):
    """
    Base task.
    """
    TAG = '133'



# coding: utf-8
# pylint: disable=C0103,W0232,C0301,W0703

# Copyright 2018 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
A new module for conversions. Complement assets.
"""

import collections
import logging
import tarfile
import tempfile
from xml.sax.saxutils import escape, unescape

import xmltodict

import marcx
import pymarc
from siskin.benchmark import timed

html_escape_table = {
    '"': "&quot;",
    "'": "&apos;"
}
html_unescape_table = {v: k for k, v in html_escape_table.items()}

logger = logging.getLogger('siskin')

def html_escape(text):
    """ Escape HTML, see also: https://wiki.python.org/moin/EscapingHtml"""
    return escape(text, html_escape_table)

def html_unescape(text):
    """ Unescape HTML, see also: https://wiki.python.org/moin/EscapingHtml"""
    return unescape(text, html_unescape_table)

@timed
def imslp_tarball_to_marc(tarball, outputfile=None, legacy_mapping=None,
                          max_failures=30):
    """
    Convert an IMSLP tarball to MARC binary without extracting it. Optionally
    write to a outputfile (filename). If outputfile is not given, write to a
    temporary location.

    Returns the location of the resulting MARC file.

    A maximum number of failed conversion can be specified with `max_failures`,
    as of 2018-04-25, there where 30 records w/o title.
    """
    if outputfile is None:
        _, outputfile = tempfile.mkstemp(prefix="siskin-")

    stats = collections.Counter()

    with open(outputfile, "wb") as output:
        writer = pymarc.MARCWriter(output)
        with tarfile.open(tarball) as tar:
            for member in tar.getmembers():
                fobj = tar.extractfile(member)
                try:
                    record = imslp_xml_to_marc(fobj.read(),
                                               legacy_mapping=legacy_mapping)
                    writer.write(record)
                except ValueError as exc:
                    logger.warn("conversion failed: %s", exc)
                    stats["failed"] += 1
                    continue
                finally:
                    fobj.close()
                    stats["processed"] += 1

        writer.close()

        if stats["failed"] > max_failures:
            raise Runtime("more than %d record failed", max_failures)

        logger.debug("%d/%d record failed/processed", stats["failed"], stats["processed"])

    return outputfile

def imslp_xml_to_marc(s, legacy_mapping=None):
    """
    Convert a string containing a single IMSLP record to MARC.

    Optionally take a legacy mapping, associating IMSLP Names with VIAF
    identifiers.  Blueprint: https://git.io/vpQPd, with one difference: We
    allow records with no subjects.

    A record w/o title is an error. We check for them, when we add the field.
    """
    dd = xmltodict.parse(s, force_list={"subject"})

    if legacy_mapping is None:
        legacy_mapping = collections.defaultdict(lambda: collections.defaultdict(str))

    record = marcx.Record(force_utf8=True)
    record.strict = False

    doc = dd["document"]

    record.leader = "     ncs  22        450 "

    identifier = doc["identifier"]["#text"]
    record.add("001", data=u"finc-15-{}".format(identifier))

    record.add("007", data="cr")
    language = doc.get("languages", "")
    record.add("008", data="130227uu20uuuuuuxx uuup%s  c" % language)

    record.add("041", a=language)
    creator = doc["creator"]["mainForm"]
    record.add("100", a=creator, e="cmp",
               _0=legacy_mapping.get(identifier, {}).get("viaf", ""))

    record.add("240", a=legacy_mapping.get(identifier, {}).get("title", ""))

    try:
        record.add("245", a=html_unescape(doc["title"]))
    except KeyError:
        raise ValueError("cannot find title: %s ..." % s[:300])

    record.add("246", a=html_unescape(doc.get("additionalTitle", "")))

    record.add("260", c=doc.get("date", ""))
    record.add("650", y=doc.get("date", ""))
    record.add("500", a=doc.get("abstract", ""))

    for689 = []

    if "subject" in doc:
        if len(doc["subject"]) == 1:
            for689.append(doc["subject"][0]["mainForm"])
        elif len(doc["subject"]) == 2:
            for689.append(doc["subject"][1]["mainForm"])
        else:
            raise ValueError("cannot handle %d subjects", len(doc["subject"]))

        record.add("590", a=for689[0].title(),
                   b=doc.get("music_arrangement_of", "").title())

        for689.append(doc.get("music_arrangement_of", ""))

        for subject in set(for689):
            record.add("689", a=subject.title())

    record.add("700", a=doc.get("contributor", {}).get("mainForm", ""), e="ctb")
    record.add("856", q="text/html", _3="Petrucci Musikbibliothek",
               u=doc["url"]["#text"])
    record.add("970", c="PN")
    record.add("980", a=identifier, b="15", c="Petrucci Musikbibliothek")
    return record




#!/usr/bin/env python
# coding: utf-8

from builtins import *

import io
import re
import sys

import marcx
import pymarc
from tqdm import tqdm

copytags = ["003", "005", "006", "008", "010", "013", "015", "016",
            "017", "020", "022", "024", "026", "028", "029", "030", "032", "033", "035",
            "040", "041", "044", "045", "049", "050", "080", "082", "084", "086", "088",
            "100", "110", "111", "130", "210", "240", "242", "243", "245", "246", "247",
            "249", "250", "255", "259", "260", "264", "270", "300", "303", "336", "337",
            "338", "340", "344", "362", "363", "382", "385", "490", "500", "501", "502",
            "505", "511", "515", "518", "520", "530", "533", "534", "535", "538", "539",
            "546", "547", "555", "581", "583", "600", "610", "611", "630", "648", "650",
            "651", "653", "655", "689", "700", "710", "711", "730", "740", "751", "770",
            "772", "773", "775", "776", "777", "780", "785", "787", "800", "810", "830",
            "850", "856", "880", "912", "924", "940", "999"]

inputfilename, outputfilename = "117_input.mrc", "117_output.mrc"

if len(sys.argv) >= 3:
    inputfilename, outputfilename = sys.argv[1:3]

# Hidden parameter, total number of records, as a hint for tqdm.
total = None if len(sys.argv) < 4 else int(sys.argv[3])

inputfile = io.open(inputfilename, "rb")
outputfile = io.open(outputfilename, "wb")

reader = pymarc.MARCReader(inputfile)

# Filtering by RVK. There are other classification schemes used in 084.2, e.g. in the matched subset:
#
#  274772 rvk
#   34325 sdnb
#   12453 ssgn
#    9641 stub
#    9222 ifzs
#    1361 sbb
#     763 dopaed
#     406 fid
#     190 z
#      42 sfb
#       6 kab
#       5 ssd
#       5 ghbs
#       3 zdbs

# A plain whitelist.
whitelist = set([
    "AN 1780",
    "AN 3900",
    "AN 3920",
    "AN 4030",
    "CV 3500",
    "DW 4000",
    "DW 4200",
    "MF 1000",
    "MF 1500",
    "NQ 2270"
])

# Extra patterns.
pattern_ms = re.compile(r"^MS.7[89][56789].*$")
pattern_ap = re.compile(r"^AP.*$")
pattern_f2 = re.compile(r"^AP.99[012].*$")  # via: filter2_relevant_for_FID.xml

# Blacklist of signatures.
blacklist = set([
    "AP 6000",
    "AP 6300",
    "AP 6400",
    "AP 6500",
    "AP 6582",
    "AP 6583",
    "AP 6586",
    "AP 6600",
    "AP 6630",
    "AP 6800",
    "AP 6930",
    "AP 7200",
    "AP 7250",
    "AP 7320",
    "AP 7337",
    "AP 7900",
    "AP 8300",
    "AP 8735",
    "AP 8786",
    "AP 9950",
    "AP 9954"
])


def filter_084a(value):
    """
    A filter helper for a single 084a value. Returns True, if the value passes.
    """
    if value in whitelist:
        return True
    if pattern_ms.match(value):
        return True
    if pattern_ap.match(value) and value not in blacklist and not pattern_f2.match(value):
        return True
    return False


# Transformation.
for oldrecord in tqdm(reader, total=total):

    newrecord = marcx.Record()

    # via: filter_DE-B170.xml
    isils = set([s for f in oldrecord.get_fields("049")
                 for s in f.get_subfields("a")])

    if "DE-B170" not in isils:
        continue

    # prüfen, ob für adlr relevante RVK-Klasse
    rvk = [s for f in oldrecord.get_fields("084")
           for s in f.get_subfields("a")]
    for value in rvk:
        if filter_084a(value):
            break
    else:
        continue

    # prüfen, ob Titel vorhanden ist
    if not oldrecord["245"]:
        continue

    for field in oldrecord.get_fields("856"):
        f856 = field if "http" in field else ""

    # leader
    newrecord.leader = "     " + oldrecord.leader[5:]

    # 001
    f001 = oldrecord["001"].data
    newrecord.add("001", data="finc-117-%s" % f001)

    # 007
    if "007" in oldrecord:
        f007 = oldrecord["007"].data
        if len(f007) < 2:
            f007 = f007 + "u"
        newrecord.add("007", data=f007)

    # Originalfelder, die ohne Änderung übernommen werden
    for tag in copytags:
        for field in oldrecord.get_fields(tag):
            newrecord.add_field(field)

    # 980
    collections = ["a", f001, "b", "117", "c",
                   "UdK Berlin", "c", "Verbundkatalog Film"]
    newrecord.add("980", subfields=collections)

    outputfile.write(newrecord.as_marc())

inputfile.close()
outputfile.close()



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301
# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Tracy Hoffmann, <tracy.hoffmann@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
KHM Koeln, refs #9269, #8391.

Configuration keys:

[khm]

scp-src = user@ftp.example.com:/home/gbi

"""

import datetime
import itertools
import os
import re
import tempfile

import luigi
from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import monthly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.benchmark import timed
from siskin.common import FTPMirror
from siskin.task import DefaultTask
from siskin.utils import iterfiles


class KHMTask(DefaultTask):
    """ Base task for KHM Koeln (sid 109) """
    TAG = '109'
    FILEPATTERN = r'.*aleph.ALL_RECS.(?P<date>[12][0-9]{3,3}[01][0-9][0123][0-9]).*(?P<no>[0-9]{1,}).tar.gz'

    def closest(self):
        return monthly(date=self.date)


class KHMDropbox(KHMTask):
    """
    Pull down content from FTP.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return Executable('rsync', message='https://rsync.samba.org/')

    def run(self):
        target = os.path.join(self.taskdir(), 'mirror')
        shellout("mkdir -p {target} && rsync {rsync_options} {src} {target}",
                 rsync_options=self.config.get('khm', 'rsync-options', fallback='-avzP'),
                 src=self.config.get('khm', 'scp-src'), target=target)

        if not os.path.exists(self.taskdir()):
            os.makedirs(self.taskdir())

        with self.output().open('w') as output:
            for path in sorted(iterfiles(target)):
                output.write_tsv(path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='filelist'), format=TSV)


class KHMLatestDate(KHMTask):
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return KHMDropbox(date=self.date)

    def run(self):
        """
        Naming schema: aleph.ALL_RECS.20170316.123655.2.tar.gz. aleph.ALL_RECS.YYYYMMDD.HHMMSS.NO.tar.gz
        """
        with self.input().open() as handle:
            with self.output().open('w') as output:
                records = sorted(handle.iter_tsv(cols=('path',)), reverse=True)
                if len(records) == 0:
                    raise RuntimeError('no files found, cannot determine latest date')
                groups = re.search(self.FILEPATTERN, records[0].path).groupdict()
                output.write_tsv(groups["date"])

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class KHMLatest(KHMTask):
    """
    Decompress and join files from the (presumably) lastest version.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    ext = luigi.Parameter(default="xml", description="mrc, xml", significant=False)

    def requires(self):
        return {
            'dropbox': KHMDropbox(date=self.date),
            'date': KHMLatestDate(date=self.date),
        }

    def run(self):
        """
        Naming schema: aleph.ALL_RECS.20170316.123655.2.tar.gz. aleph.ALL_RECS.YYYYMMDD.HHMMSS.NO.tar.gz

        Ugly XML merge.
        """
        with self.input().get('date').open() as handle:
            latest_date = handle.read().strip()

        _, stopover = tempfile.mkstemp(prefix='siskin-')

        shellout(""" echo '<?xml version = "1.0" encoding = "UTF-8"?>
                           <collection xmlns="http://www.loc.gov/MARC21/slim"
                                       xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                                       xsi:schemaLocation="http://www.loc.gov/MARC21/slim http://www.loc.gov/standards/marcxml/schema/MARC21slim.xsd">
            ' >> {stopover} """, stopover=stopover, preserve_whitespace=True)

        with self.input().get('dropbox').open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                groups = re.search(self.FILEPATTERN, row.path).groupdict()
                if groups["date"] != latest_date:
                    continue
                shellout("tar xOf {input} | xmlcutty -path /OAI-PMH/ListRecords/record/metadata/record >> {stopover}", input=row.path, stopover=stopover)

        shellout(""" echo '</collection>' >> {stopover} """, stopover=stopover)
        output = shellout(""" xmllint --format {input} |
                              grep -v '<controlfield tag="LDR">' |
                              grep -v '<controlfield tag="FMT">' > {output} """, input=stopover)
        if self.ext == "mrc":
            output = shellout(""" yaz-marcdump -i marcxml -o marc {input} > {output}""", input=output)

        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext=self.ext))


class KHMIntermediateSchema(KHMTask):
    """
    Convert to intermediate schema via metafacture.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return KHMLatest(date=self.date)

    def run(self):
        mapdir = 'file:///%s' % self.assets("maps/")
        output = shellout("""flux.sh {flux} in={input} MAP_DIR={mapdir} > {output}""",
                          flux=self.assets("109/109.flux"), mapdir=mapdir, input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj'))



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,R0904,E1103,C0301

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Medienwissenschaft, Rezensionen, Reviews. Archiv, Marburg, refs #5486, #11005.

Previously:

* http://archiv.ub.uni-marburg.de/ep/0002/oai

Now (2018-04-23):

* http://archiv.ub.uni-marburg.de/ubfind/OAI/Server

"""

import datetime
import json
import os
import tempfile

import luigi
import xmltodict
from gluish.format import Gzip
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask


class MarburgTask(DefaultTask):
    """ Base task for Marburg. """
    TAG = '73'

    def closest(self):
        return weekly(self.date)


class MarburgCombine(MarburgTask):
    """
    Harvest and combine into a single file.

    XXX: NLM format has been discontinued as of 2018-01-01, refs #5486.
    XXX: Maybe datacite.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='nlm')

    def run(self):
        endpoint = "http://archiv.ub.uni-marburg.de/ubfind/OAI/Server"
        shellout("metha-sync -format nlm {endpoint}", endpoint=endpoint)
        output = shellout("metha-cat -format nlm {endpoint} > {output}", endpoint=endpoint)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(self.path())


class MarburgMarc(MarburgTask):
    """
    Convert XML to Marc.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='datacite')

    def requires(self):
        return MarburgCombine(format=self.format)

    def run(self):
        output = shellout("python {script} {input} > {output}",
                          script=self.assets("73/73_marcbinary.py"), input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(self.path(ext='mrc'))


class MarburgJSON(MarburgTask):
    """
    Convert XML to JSON in one go. Preparation, so we can use jq(1).
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MarburgCombine(date=self.date)

    def run(self):
        with self.input().open() as handle:
            with self.output().open('w') as output:
                json.dump(xmltodict.parse(handle.read()), output)

    def output(self):
        return luigi.LocalTarget(self.path(ext='json'))


class MarburgIntermediateSchema(MarburgTask):
    """
    Convert to intermediate schema.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return MarburgJSON(date=self.date)

    def run(self):
        output = shellout(""" cat {input} | \
                              jq '.Records.Record[]|.metadata.article' | \
                              jq -f {filter} -cr | gzip -c > {output}""",
                          input=self.input().path, filter=self.assets('73/filter.jq'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(self.path(ext='ldj.gz'), format=Gzip)


class MarburgExport(MarburgTask):
    """
    Attach and export.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3')

    def requires(self):
        return {
            'data': MarburgIntermediateSchema(date=self.date),
            'config': AMSLFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout("span-tag -c {config} <(gunzip -c {input}) | span-export -o {format} > {output}",
                          config=self.input().get("config").path, input=self.input().get("data").path,
                          format=self.format)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(self.path(ext='ldj'))


class MarburgCombineNext(MarburgTask):
    """
    Harvest and combine into a single file.

    Adjust format, refs #5486.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='nlm')

    def run(self):
        endpoint = "http://archiv.ub.uni-marburg.de/ubfind/OAI/Server"
        output = shellout("oaicrawl -w 8 -verbose -f oai_dc {endpoint} > {output}", endpoint=endpoint)
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout("echo '<records>' >> {output}", output=stopover)
        shellout("xmlcutty -path /OAI-PMH/GetRecord/record < {input} >> {output}", input=output, output=stopover)
        shellout("echo '</records>' >> {output}", output=stopover)
        luigi.LocalTarget(stopover).move(self.output().path)
        # Remove raw data.
        os.remove(output)

    def output(self):
        return luigi.LocalTarget(self.path())



# coding: utf-8
#
# Copyright 2018 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>#                   
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Cambridge Open Access eBooks, refs #13120.

Download data manually from https://www.cambridge.org/core/services/librarians/marc-records.

Configuration:

[coaeb]

input = /path/to/mrc (ask RS)
"""

import luigi
from gluish.utils import shellout

from siskin.task import DefaultTask


class COAEBTask(DefaultTask):
    """
    This task inherits functionality from `siskin.task.DefaultTask`.
    """
    TAG = '161'


class COAEBMARC(COAEBTask):
    """ Convert BinaryMARC to BinaryFincMarc """

    def run(self):
        output = shellout("""python {script} {input} {output}""",
                          script=self.assets("161/161_marcbinary.py"),
                          input=self.config.get("coaeb", "input"))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='fincmarc.mrc'))




# coding: utf-8
# pylint: disable=W0232,E1101,R0904,E1103,C0301

# Copyright 2017 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Robert Schenk, <robert.schenk@uni-leipzig.de>
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
#

"""
B3Kat, #8697.
"""

import datetime
import os
import tempfile

import luigi
from gluish.format import TSV
from gluish.intervals import semiyearly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.task import DefaultTask


class B3KatTask(DefaultTask):
    """ Base task for B3Kat. """
    TAG = 'b3kat'

    def closest(self):
        """ Find date on this site: https://www.bib-bvb.de/web/b3kat/open-data """
        return semiyearly(date=self.date)


class B3KatLinks(B3KatTask):
    """
    A list of latest links. Currently grepped from https://www.bib-bvb.de/web/b3kat/open-data.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def run(self):
        output = shellout("""
            curl --fail -s "https://www.bib-bvb.de/web/b3kat/open-data" |
            grep -Eo "/OpenData/b3kat_export_[0-9]{{4,4}}_[0-9]{{1,2}}_teil[0-9]{{1,2}}.xml.gz" |
            awk '{{print "https://www.bib-bvb.de"$0 }}' > {output} """)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class B3KatDownload(B3KatTask):
    """
    Download snapshot. Output is a single (large) MARC binary file. Strip
    non-sorting characters (U+0098, U+009C) from raw data. XXX: Single core.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return B3KatLinks(date=self.date)

    def run(self):
        """
        XXX: wrapped sed ineffective.
        """
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for i, row in enumerate(handle.iter_tsv(cols=('url',)), start=1):
                downloaded = shellout("""curl -sL --fail "{url}" > {output} """, url=row.url)
                cleaned = shellout(r"""CHARS=$(python -c 'print u"\u0098\u009C".encode("utf8")');
                                       unpigz -c {input} | sed 's/['"$CHARS"']//g' > {output} """, input=downloaded)
                output = shellout("""yaz-marcdump -i marcxml -o marc "{input}" > {output}""", input=cleaned)
                shellout("cat {input} >> {stopover}",
                         input=output, stopover=stopover)
                try:
                    os.remove(downloaded)
                    os.remove(cleaned)
                    os.remove(output)
                except OSError as err:
                    self.logger.error(err)

        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='mrc'))


class B3KatMARCXML(B3KatTask):
    """
    Convert binary MARC to MARCXML.
    """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return B3KatDownload(date=self.date)

    def run(self):
        output = shellout("yaz-marcdump -i marc -o marcxml {input} | pigz -c > {output}",
                          input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml.gz'))



#!/usr/bin/env python
# coding: utf-8

import io
import re
import sys
import json
import base64

import marcx
import xmltodict


colormap = {
    "black and white": "Schwarzweiß",
    "black & white": "Schwarzweiß",
    "B&W": "Schwarzweiß",
    "color": "Farbe",
    "b&w": "Schwarzweiß",
    "Color": "Farbe",
    "Mix": "Schwarzweiß / Farbe",  
    "B&W (tinded)": "Schwarzweiß (koloriert)",
    "b/w": "Schwarzweiß",
    "c": "Farbe",
    "tinted": "Schwarzweiß (koloriert)",
    "bw": "Schwarzweiß",
    "C": "Farbe",  
    "B&W/C": "Schwarzweiß / Farbe",
    "tinted B&W": "Schwarzweiß (koloriert)",
    "B&W (tinted)": "Schwarzweiß (koloriert)",
    "sepia": "Farbe",
    "B/W": "Farbe",
    "color/B&W": "Schwarzweiß / Farbe",
    "b&w/color": "Schwarzweiß / Farbe",
    "Black and White": "Schwarzweiß",
    "B&w": "Schwarzweiß"
}

soundmap = {
    "black and white": "Schwarzweiß",
    "sound": "Ton",
    "silent": "Stumm",
    "No": "Stumm",
    "Yes": "Ton",
    "Sd": "Ton",
    "si": "Stumm",
    "Si": "Stumm",  
    "silen": "Stumm",  
    "sd": "Ton",
    "SD": "Ton"
}


def get_field(tag):
    try:
        return jsonrecord[tag]        
    except:
        return ""


inputfilename = "153_input.ldj" 
outputfilename = "153_output.mrc"

if len(sys.argv) == 3:
    inputfilename, outputfilename = sys.argv[1:]

inputfile = open(inputfilename, "r")
outputfile = open(outputfilename, "wb")

for line in inputfile:

    jsonobject = json.loads(line)
    jsonrecord = jsonobject["metadata"]    

    marcrecord = marcx.Record(force_utf8=True)

    # Leader
    marcrecord.leader = "     cam  22        4500"

    # Identifier
    try:
        f001 = jsonrecord["identifier"]       
    except:
        continue
    
    f001 = f001.encode("utf-8")
    f001 = base64.b64encode(f001)
    f001 = f001.decode("ascii")
    f001 = f001.rstrip("=")
    marcrecord.add("001", data="finc-153-" + f001)        
   
    # Format
    marcrecord.add("007", data="cr")

    # Urheber
    f110a = get_field("creator")
    if f110a != "" and f110a != "Unknown":
        marcrecord.add("110", a=f110a)
    
    # Hauptitel
    f245a = get_field("title")
    marcrecord.add("245", a=f245a)

    # gestattet leere Felder, solange der Titel und der Identifier vorhanden sind
    marcrecord.strict = False

    # Filmstudio
    f260b = get_field("publisher")        
  
    # Erscheinungsjahr
    f260c = get_field("date")
    if f260c != "":
        regexp = re.search("(\d\d\d\d)", f260c)
        if regexp:
            f260c = regexp.group(1)
    
    # Verlag    
    if f260b != "" and f260b == f110a:  
        f260b = ""  # verhindert, dass Körperschaft und Verlag nebeneinander im Katalog angezeigt werden, wenn beide identisch sind

    if f260b != "" and f260c != "":
        f260b = f260b + ", "  # ergänzt das Trennzeichen zwischen Produktionsfirma und Jahr, wenn beides vorhanden   
    publisher = ["b", f260b, "c", f260c]
    marcrecord.add("260", subfields=publisher)  

    # Spielzeit
    runtime = get_field("runtime")
   
    # Bild
    color_old = get_field("color")
    if color_old != "":
        color = colormap.get(color_old, "")
    else:
        color = ""
    if color == "" and color_old != "":
        print("Die Farbe %s wurde in der Colormap nicht gefunden" % color_old)
  
    # Ton
    sound_old = get_field("sound")
    if sound_old != "":
        sound = soundmap.get(sound_old, "")
    else:
        sound = ""
    if sound == "" and sound_old != "":
        print("Der Sound %s wurde in der Soundmap nicht gefunden" % sound_old)
   
    if color != "" and sound != "":
        f300b = color + " + " + sound
    elif color != "":
        f300b = color
    elif sound != "":
        f300b = sound
    else:
        f300b = ""

    if runtime != "":      
        runtime = runtime.lstrip("00:")
        runtime = runtime.lstrip("0:")
        runtime = " (" + runtime
        if "min." not in runtime:
            runtime = runtime + " min.)"
        else:
            runtime = runtime + ")"
        f300a = f300b + runtime
    else:
        f300a = f300b  
 
    marcrecord.add("300", a=f300a)

    # Spielzeit (extra MARC-Feld)
    f306a = get_field("runtime")
    if f306a != "":
        f306a = f306a.lstrip("00:")
        f306a = f306a.lstrip("0:")
        if "min" not in f306a:
            f306a = f306a + " min."
    marcrecord.add("306", a=f306a)   

    # Soundformat (extra MARC-Feld)
    marcrecord.add("344", a=sound)  
    
    # Bildformat (extra -MARC-Feld)
    marcrecord.add("346", a=color)

    # Annotation 
    f520a = get_field("description")
    marcrecord.add("520", a=f520a)

    # Schlagwörter
    subjects = jsonrecord.get("subject", "")    
    if subjects != "":
        if isinstance(subjects, list):
            for subject in subjects:              
                subject = subject.title()
                marcrecord.add("650", a=subject)            
        else:
            if subjects != "need keyword":
                subject = subjects.title()             
                subjects = subject.split(";")
                if isinstance(subjects, list):
                    for subject in subjects:              
                        subject = subject.title()
                        marcrecord.add("650", a=subject)
                else:
                    marcrecord.add("650", a=subject)


    # hebt die Erlaubnis für leere Felder wieder auf
    marcrecord.strict = True         
    
    # Link zur Ressource
   
    f856u = get_field("identifier")
    marcrecord.add("856", q="text/html", _3="Link zur Ressource", u="https://archive.org/details/" + f856u)

    # Medienform
    marcrecord.add("935", b="cofz", c="vide")

    # Kollektion   
    marcrecord.add("980", a=f001, b="153", c="Internet Archive / Prelinger")

    outputfile.write(marcrecord.as_marc())

inputfile.close()
outputfile.close()


# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
DeGruyter tasks.

[degruyter]

ftp-host = host.name
ftp-username = username
ftp-password = password
ftp-path = /
ftp-pattern = some*glob*pattern.zip

"""

import datetime
import tempfile

import luigi
from gluish.format import TSV
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.benchmark import timed
from siskin.common import Executable, FTPMirror
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask


class DegruyterTask(DefaultTask):
    """
    De Gruyter is a scholarly publishing house specializing in academic literature.

    Each file on the FTP server comes with an epoch timestamp, which seems to be the
    same for a single shipment:

        ...
        zrgra.2010.127.issue-1-1496849175274.zip
        zrs.2012.4.issue-1-1496849175274.zip
        ...

    The fileset we use is based on this timestamp and can be set with TIMESTAMP.

        $ ...80420a2/SSH $ ls -1 | egrep -o '1[0-9]{4,}' | sort | uniq -c | sort -nr
        8878 1497254827738
        8878 1496849175274
        8023 1423571177743
        1041 1424176025682
           2 1478602063508

    """
    TAG = 'degruyter'
    TIMESTAMP = '1497254827738'  # TODO(miku): partial updates are possible and missed by this.

    def closest(self):
        return datetime.date(2017, 8, 1)


class DegruyterPaths(DegruyterTask):
    """ A list of Degruyter ile paths (via FTP). """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        host = self.config.get('degruyter', 'ftp-host')
        username = self.config.get('degruyter', 'ftp-username')
        password = self.config.get('degruyter', 'ftp-password')
        base = self.config.get('degruyter', 'ftp-path')
        pattern = self.config.get('degruyter', 'ftp-pattern')
        exclude_glob = self.config.get('degruyter', 'ftp-exclude-glob', fallback='')
        return FTPMirror(host=host, username=username, password=password,
                         base=base, pattern=pattern, exclude_glob=exclude_glob)

    @timed
    def run(self):
        self.input().move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="filelist"), format=TSV)


class DegruyterXML(DegruyterTask):
    """
    Single file version.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    group = luigi.Parameter(default='SSH', description='main subdirectory')
    ts = luigi.Parameter(default=DegruyterTask.TIMESTAMP)

    def requires(self):
        return DegruyterPaths(date=self.date)

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        with self.input().open() as handle:
            for row in handle.iter_tsv(cols=('path',)):
                if not '/%s/' % self.group in row.path:
                    continue
                if '-%s.zip' % self.ts not in row.path:
                    continue
                shellout(r"unzip -p {path} \*.xml 2> /dev/null >> {output}", output=stopover, path=row.path,
                         ignoremap={1: 'OK', 9: 'skip corrupt file'})
        luigi.LocalTarget(stopover).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='xml'), format=TSV)


class DegruyterIntermediateSchema(DegruyterTask):
    """ Convert to intermediate format via span. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'span': Executable(name='span-import', message='http://git.io/vI8NV'),
                'file': DegruyterXML(date=self.date)}

    @timed
    def run(self):
        output = shellout("span-import -i degruyter {input} | pigz -c > {output}", input=self.input().get('file').path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class DegruyterExport(DegruyterTask):
    """
    A SOLR-importable format.
    """
    date = luigi.DateParameter(default=datetime.date.today())
    format = luigi.Parameter(default='solr5vu3')

    def requires(self):
        return {
            'file': DegruyterIntermediateSchema(date=self.date),
            'config': AMSLFilterConfig(date=self.date)
        }

    @timed
    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path, input=self.input().get('file').path)
        output = shellout("span-export -o {format} <(unpigz -c {input}) | pigz -c > {output}", input=output, format=self.format)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        extensions = {
            'solr5vu3': 'ldj.gz',
            'formeta': 'form.gz',
        }
        return luigi.LocalTarget(path=self.path(ext=extensions.get(self.format, 'gz')))


class DegruyterISSNList(DegruyterTask):
    """ List of ISSNs. """

    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return DegruyterIntermediateSchema(date=self.date)

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout("""jq -r '.["rft.issn"][]?' <(unpigz -c {input}) 2> /dev/null >> {output} """, input=self.input().path, output=stopover)
        shellout("""jq -r '.["rft.eissn"][]?' <(unpigz -c {input}) 2> /dev/null >> {output} """, input=self.input().path, output=stopover)
        output = shellout("""sort -u {input} > {output} """, input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)


class DegruyterDOIList(DegruyterTask):
    """ A list of Degruyter DOIs. """
    date = ClosestDateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': DegruyterIntermediateSchema(date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    @timed
    def run(self):
        _, stopover = tempfile.mkstemp(prefix='siskin-')
        shellout("""jq -r '.doi' <(unpigz -c {input}) | grep -v "null" | grep -o "10.*" 2> /dev/null > {output} """,
                 input=self.input().get('input').path, output=stopover)
        output = shellout("""sort -u {input} > {output} """, input=stopover)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=C0301,E1101

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Config:

[core]

metha-dir = /path/to/dir

[thieme]

oai = https://example.com/oai/provider
"""

import datetime
import tempfile

import luigi
from gluish.common import Executable
from gluish.format import TSV
from gluish.intervals import weekly
from gluish.parameter import ClosestDateParameter
from gluish.utils import shellout

from siskin.decorator import deprecated
from siskin.sources.amsl import AMSLFilterConfig
from siskin.task import DefaultTask


class ThiemeTask(DefaultTask):
    """ Thieme connect. """
    TAG = '60'

    def closest(self):
        return weekly(date=self.date)


class ThiemeCombine(ThiemeTask):
    """ Combine files."""

    date = ClosestDateParameter(default=datetime.date.today())
    prefix = luigi.Parameter(default="nlm")
    set = luigi.Parameter(default='journalarticles')

    def requires(self):
        return Executable(name='metha-sync', message='https://github.com/miku/metha')

    def run(self):
        url = self.config.get('thieme', 'oai')
        shellout("METHA_DIR={dir} metha-sync -set {set} -format {prefix} {url}",
                 set=self.set, prefix=self.prefix, url=url, dir=self.config.get('core', 'metha-dir'))
        output = shellout("METHA_DIR={dir} metha-cat -format {prefix} {url} | pigz -c > {output}",
                          prefix=self.prefix, url=url, dir=self.config.get('core', 'metha-dir'))
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="xml.gz"))


class ThiemeIntermediateSchema(ThiemeTask):
    """
    Conv-oo-do-ert.
    """

    date = ClosestDateParameter(default=datetime.date.today())
    set = luigi.Parameter(default='journalarticles')

    def requires(self):
        return ThiemeCombine(date=self.date, prefix='nlm', set=self.set)

    def run(self):
        output = shellout(
            "span-import -i thieme-nlm <(unpigz -c {input}) | pigz -c > {output}", input=self.input().path)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext="ldj.gz"))


class ThiemeExport(ThiemeTask):
    """
    Create something solr importable. Attach a single ISIL to all records.

    Ad-hoc task, with all records attached.
    """
    date = ClosestDateParameter(default=datetime.date.today())
    set = luigi.Parameter(default='journalarticles')
    version = luigi.Parameter(
        default='solr5vu3', description='export JSON flavors, e.g.: solr4vu13v{1,10}, solr5vu3v11')

    def requires(self):
        return {
            'is': ThiemeIntermediateSchema(date=self.date, set=self.set),
            'config': AMSLFilterConfig(date=self.date),
        }

    def run(self):
        output = shellout("span-tag -c {config} <(unpigz -c {input}) | pigz -c > {output}",
                          config=self.input().get('config').path, input=self.input().get('is').path)
        output = shellout(
            "span-export -o {version} <(unpigz -c {input}) | pigz -c > {output}", input=output, version=self.version)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(ext='ldj.gz'))


class ThiemeISSNList(ThiemeTask):
    """
    A list of Elsevier journals DOIs.
    """
    date = luigi.DateParameter(default=datetime.date.today())

    def requires(self):
        return {'input': ThiemeIntermediateSchema(date=self.date),
                'jq': Executable(name='jq', message='https://github.com/stedolan/jq')}

    def run(self):
        _, output = tempfile.mkstemp(prefix='siskin-')
        shellout("""jq -c -r '.["rft.issn"][]?' <(unpigz -c {input}) >> {output} """, input=self.input(
        ).get('input').path, output=output)
        shellout("""jq -c -r '.["rft.eissn"][]?' <(unpigz -c {input}) >> {output} """, input=self.input(
        ).get('input').path, output=output)
        output = shellout("""sort -u {input} > {output} """, input=output)
        luigi.LocalTarget(output).move(self.output().path)

    def output(self):
        return luigi.LocalTarget(path=self.path(), format=TSV)



# coding: utf-8
# pylint: disable=F0401,C0111,W0232,E1101,E1103,C0301,C0103,W0614,W0401,E0202

# Copyright 2015 by Leipzig University Library, http://ub.uni-leipzig.de
#                   The Finc Authors, http://finc.info
#                   Martin Czygan, <martin.czygan@uni-leipzig.de>
#
# This file is part of some open source application.
#
# Some open source application is free software: you can redistribute
# it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Some open source application is distributed in the hope that it will
# be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Various utilities.
"""

from __future__ import print_function

import errno
import hashlib
import itertools
import json
import logging
import operator
import os
import random
import re
import string
import sys
import tempfile
from builtins import map, object, range, zip

import bs4
import luigi
import requests
from dateutil import relativedelta
from future import standard_library
from six.moves.urllib.parse import urlparse

import backoff
from siskin import __version__

standard_library.install_aliases()


logger = logging.getLogger('siskin')


class SetEncoder(json.JSONEncoder):
    """
    Helper to encode python sets into JSON lists.
    """

    def default(self, obj):
        """ Decorate call to standard implementation. """
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


def date_range(start_date, end_date, increment, period):
    """
    Generate `date` objects between `start_date` and `end_date` in `increment`
    `period` intervals.
    """
    result = []
    nxt = start_date
    delta = relativedelta.relativedelta(**{period: increment})
    while nxt <= end_date:
        result.append(nxt)
        nxt += delta
    return result


def iterfiles(directory='.', fun=None):
    """
    Shortcut for os.walk, yield paths below a directory, optionally filter the
    paths by function given in `fun`.
    """
    if fun is None:
        fun = lambda path: True
    for root, _, files in os.walk(directory):
        for f in files:
            path = os.path.join(root, f)
            if fun(path):
                yield path


def random_string(length=16):
    """
    Return a random string(upper and lowercase letters) of length `length`,
    defaults to 16.
    """
    return ''.join(random.choice(string.ascii_letters) for _ in range(length))


def pairwise(obj):
    """
    Iterator over a iterable in steps of two.
    """
    iterable = iter(obj)
    return zip(iterable, iterable)


def nwise(iterable, n=2):
    """
    Generalized: func: `pairwise`. Split an iterable after every `n` items.
    """
    i = iter(iterable)
    piece = tuple(itertools.islice(i, n))
    while piece:
        yield piece
        piece = tuple(itertools.islice(i, n))


def dictcheck(obj, contains=None, absent=None):
    """
    Check if a dictionary contains values for the keys given in `contains` and at
    the same time it does not contain values for keys, which are given in
    `absent`.

    >>> dictcheck({"name": "x"}, contains=["name"])
    True

    >>> dictcheck({"name": "x"}, contains=["name"], absent=["key"])
    True

    >>> dictcheck({"name": "x", "key": "1234"}, contains=["name"], absent=["key"])
    False

    >>> dictcheck({"key": None}, absent=["key"])
    True

    >>> dictcheck({}, absent=["key"])
    True

    This is a helper for AMSL-API cases (https://git.io/vpHgS).
    """
    if not isinstance(obj, dict):
        raise ValueError('dictionary only')
    if contains is None:
        contains = []
    if absent is None:
        absent = []

    for key in contains:
        if key not in obj:
            return False
        if not bool(operator.itemgetter(key)(obj)):
            return False

    for key in absent:
        if key not in obj:
            continue
        if bool(operator.itemgetter(key)(obj)):
            return False

    return True


def get_task_import_cache():
    """
    Load `taskname: modulename` mappings from dictionary. Return a tuple containing
    the dictionary and the path to the cache file.

    The command line entry points (e.g. taskdo and friends) need to import all
    modules in order to find a task. The import process actually takes a while,
    so the import cache shortens the startup time of the command line tools by
    only importing the module the given task is in.

    It is save to remove the path returned by `taskimportcache` at any time.
    """
    task_import_cache = None
    path = os.path.join(tempfile.gettempdir(),
                        'siskin_task_import_cache_%s' % __version__)
    if not os.path.exists(path):
        logger.debug("creating task import cache at %s", path)
        from siskin.cacheutils import _write_task_import_cache
        _write_task_import_cache(path)

    with open(path) as handle:
        try:
            task_import_cache = json.load(handle)
        except Exception as err:
            print("failed to load task import cache, remove %s, then try again" %
                  path, file=sys.stderr)
            sys.exit(1)

    return task_import_cache, path


def load_set_from_target(target):
    """
    Given a luigi.LocalTarget, load each line of the file into a set.
    """
    s = set()
    with target.open() as handle:
        for line in handle:
            s.add(line.strip())
    return s


def load_set_from_file(filename, func=lambda v: v):
    """
    Given a filename, load each non-empty line into a set.
    """
    s = set()
    with open(filename) as handle:
        for line in map(str.strip, handle):
            if not line:
                continue
            s.add(func(line))
    return s


class URLCache(object):
    """
    A simple URL content cache. Stores everything on the filesystem. Content
    is first written to a temporary file and then renamed. With concurrent
    requests for the same URL, the last one wins. Raises exception on any HTTP
    status >= 400. Retries supported.

    It is not very efficient, as it creates lots of directories.
    > 396140 directories, 334024 files ... ...

    >>> cache = URLCache()
    >>> cache.get_cache_file("https://www.google.com")
    /tmp/ef/7e/fc/ef7efc9839c3ee036f023e9635bc3b056d6ee2d

    >>> cache.is_cached("https://www.google.com")
    False

    >>> page = cache.get("https://www.google.com")
    >>> page[:15]
    '<!doctype html>'

    >>> cache.is_cached("https://www.google.com")
    True

    >>> page = cache.get("https://www.google.com", force=True) # Always download.
    """

    def __init__(self, directory=None, max_tries=12):
        """
        If `directory` is not explictly given, all files will be stored under
        the temporary directory. Requests can be retried, if they resulted in
        a non 200 HTTP status code. The server might send a HTTP 500 (Internal
        Server Error), even if it really is a HTTP 503 (Service Unavailable).
        We therefore treat HTTP 500 errors as something to retry on,
        at most `max_tries` times.
        """
        self.directory = directory or tempfile.gettempdir()
        self.sess = requests.session()
        self.max_tries = max_tries

    def get_cache_file(self, url):
        """
        Return the cache file path for a URL. This will - as a side effect -
        create the parent directories, if necessary.
        """
        digest = hashlib.sha1(url).hexdigest()
        d0, d1, d2 = digest[:2], digest[2:4], digest[4:6]
        path = os.path.join(self.directory, d0, d1, d2)

        if not os.path.exists(path):
            try:
                os.makedirs(path)
            except OSError as e:
                if e.errno == errno.EEXIST:
                    pass
                else:
                    raise
        return os.path.join(path, digest)

    def is_cached(self, url):
        return os.path.exists(self.get_cache_file(url))

    def get(self, url, force=False):
        """
        Return URL, either from cache or the web.
        """

        @backoff.on_exception(backoff.expo, RuntimeError, max_tries=self.max_tries)
        def fetch(url):
            """
            Nested function, so we can configure number of retries.
            """
            r = self.sess.get(url, timeout=600)
            if r.status_code >= 400:
                raise RuntimeError('%s on %s' % (r.status_code, url))
            with tempfile.NamedTemporaryFile(delete=False) as output:
                output.write(r.text.encode('utf-8'))
            os.rename(output.name, self.get_cache_file(url))

        if not self.is_cached(url) or force is True:
            fetch(url)

        with open(self.get_cache_file(url)) as handle:
            return handle.read()

def scrape_html_listing(url, with_head=False):
    """
    Given a URL to a webpage, try to create a JSON representation of files on
    the page.

        >>> scrape_html_listing("https://www.colorado.edu/physics/phys1120/phys1120_fa09/LectureNotes/")
        ['https://www.colorado.edu/physics/phys1120/phys1120_fa09/LectureNotes/BFieldPictures.doc',
         'https://www.colorado.edu/physics/phys1120/phys1120_fa09/LectureNotes/BFieldPictures.pdf',
         ...
        ]

    Will fail if the request fails. If parsing fails, return an empty list.
    Optionally, only include links in the list which return something ok on
    HTTP HEAD (might take a while).
    """
    filelike_p = re.compile(r"[0-9-_\w.]*[.][a-z0-9]{2,4}")

    # Find base url to prepend on relative links.
    pr = urlparse(url)
    pr = pr._replace(path=os.path.dirname(pr.path))
    pr = pr._replace(params="")
    pr = pr._replace(query="")
    pr = pr._replace(fragment="")
    baseurl = pr.geturl()

    r = requests.get(url)
    if r.status_code >= 400:
        raise RuntimeError("fetch failed with %s: %s", r.status_code, url)

    soup = bs4.BeautifulSoup(r.text)
    links = set()

    for a in soup.find_all("a"):
        match = filelike_p.search(a.get("href"))
        if not match:
            continue

        # This might be an absolute link or just a filename.
        link = match.group()
        if not link.startswith("http") and not link.startswith("/"):
            link = os.path.join(baseurl, link)
        if not with_head or requests.head(link).status_code < 400:
            links.add(link)

    return sorted(links)
